from __future__ import annotations
from ..function import *
from .types import *
import math

class Camera3D:
    position: Vec3D
    rotation: Vec3D
    fov: float
    draw_distance: float
    aspect: Vec2D
    def __init__(self, width: int, height: int, fov: float, distance: float = 100):
        self.position = Vec3D(0, 0, 0)
        self.rotation = Vec3D(0, 0, 0)

        self.fov = fov
        self.draw_distance = distance

        self.aspect = Vec2D(width, height)
    
    def look_at(self, target: Vec3D):
        dx = target.x - self.position.x
        dy = target.y - self.position.y
        dz = target.z - self.position.z

        yaw = math.atan2(dz, dx)

        dist = math.sqrt(dx * dx + dz * dz)

        pitch = -math.atan2(dy, dist)

        self.rotation.x = -math.degrees(pitch)
        self.rotation.y = math.degrees(yaw)
        self.rotation.z = 0

    def get_forward(self):
        pitch = math.radians(self.rotation.x)
        yaw = math.radians(self.rotation.y)
        return Vec3D(
            math.cos(pitch) * math.cos(yaw),
            math.sin(pitch),
            math.cos(pitch) * math.sin(yaw)
        ).normalized()
    
    def get_right(self):
        pitch = math.radians(self.rotation.x)
        yaw = math.radians(self.rotation.y + 90)
        return Vec3D(
            math.cos(pitch) * math.cos(yaw), 
            0,
            math.cos(pitch) * math.sin(yaw)
        ).normalized()

class Space3D:
    _id: int
    def __init__(self):
        self._id = lib.create_space3d()

    def clear(self):
        lib.clear_space3d(self._id)
    
    def draw(self, window: Window, camera: Camera3D):
        lib.draw_camera_3d_win(window._id, self._id,
                               camera.position.x, camera.position.y, camera.position.z,
                               camera.rotation.x, camera.rotation.y, camera.rotation.z,
                               int(camera.aspect.x), int(camera.aspect.y), camera.fov, camera.draw_distance)



class draw3d:
    class light:
        @staticmethod
        def ambient(space: Space3D, power: float,
                    color: RGBA = RGBA(255, 255, 255, 255)):
            lib.draw_ambient_light_3d(
                space._id,
                power,
                color.r / 255.0,
                color.g / 255.0,
                color.b / 255.0,
                color.a / 255.0
            )

        @staticmethod
        def point(space: Space3D, position: Vec3D, power: float,
                    color: RGBA = RGBA(255, 255, 255, 255)):
            lib.draw_point_light_3d(
                space._id,
                position.x,
                position.y,
                position.z,
                power,
                color.r / 255.0,
                color.g / 255.0,
                color.b / 255.0,
                color.a / 255.0
            )

    @staticmethod
    def cube(space: Space3D,
             position: Vec3D,
             size: Vec3D,
             rotation: Vec3D = Vec3D(0, 0, 0),
             color: RGBA = RGBA(255, 255, 255, 255)):
        lib.draw_cube_3d(
            space._id,
            position.x, position.y, position.z,
            size.x, size.y, size.z,
            math.radians(rotation.x),
            math.radians(rotation.y),
            math.radians(rotation.z),
            color.r / 255.0,
            color.g / 255.0,
            color.b / 255.0,
            color.a / 255.0
        )


from __future__ import annotations
from ..function import *
import math
from .types import *

class draw:
    @staticmethod
    def rect(window: Surface, tex: Texture, rect: Rect, color: RGBA=RGBA(255,255,255,255)):
        if window.type == SurfaceType.WINDOW:
            lib.draw_rect_win(window._id, rect.x, rect.y, rect.w, rect.h, color.r / 255.0, color.g / 255.0, color.b / 255.0, color.a / 255.0)
        elif window.type == SurfaceType.TEXTURE:
            lib.draw_rect_tex(window._id, rect.x, rect.y, rect.w, rect.h, color.r / 255.0, color.g / 255.0, color.b / 255.0, color.a / 255.0)

    @staticmethod   
    def texture(window: Surface, tex: Texture, rect: Rect, color: RGBA=RGBA(255,255,255,255), flipX=False, flipY=False):
        if window.type == SurfaceType.WINDOW:
            lib.draw_texture_win(window._id, tex._id, rect.x, rect.y, rect.w, rect.h, color.r / 255.0, color.g / 255.0, color.b / 255.0, color.a / 255.0, flipX, flipY)
        elif window.type == SurfaceType.TEXTURE:
            lib.draw_texture_tex(window._id, tex._id, rect.x, rect.y, rect.w, rect.h, color.r / 255.0, color.g / 255.0, color.b / 255.0, color.a / 255.0, flipX, flipY)

    @staticmethod
    def text(window: Surface, font: Font, text: str, x: float, y: float, size: float, color: RGBA = RGBA(255,255,255,255)):
        if window.type == SurfaceType.WINDOW:
            lib.draw_text_win(window._id, font._id, text.encode("utf-8"), x, y, size / 1500, color.r / 255.0, color.g / 255.0, color.b / 255.0, color.a / 255.0)
        elif window.type == SurfaceType.TEXTURE:
            lib.draw_text_tex(window._id, font._id, text.encode("utf-8"), x, y, size / 1500, color.r / 255.0, color.g / 255.0, color.b / 255.0, color.a / 255.0)



from __future__ import annotations
from .function import *
import math
from enum import Enum
from .src.types import *
from .src.draw import *
from .src._3d import *

def quit():
    lib.finalization()

def now_time():
    return lib.get_time()

_last_time = 0
_delta_time = 0
def start_frame():
    global _last_time, _delta_time
    current_time = now_time()
    _delta_time = current_time - _last_time
    _last_time = current_time
    lib.start_frame()

def delta_time():
    return _delta_time

def load_texture(*args):
    if len(args) == 1:
        # Only the path was passed
        tex = Texture(0, 0)
        path = args[0]
    elif len(args) == 2:
        tex = args[0]
        path = args[1]
    else:
        raise TypeError(
            "load_texture(str) or load_texture(Texture, str)"
        )
    if isinstance(tex, Texture) and isinstance(path, str):
        lib.load_texture(tex._id, path.encode("utf-8"), False)
    return tex

def load_image(*args):
    if len(args) == 1:
        # Only the path was passed
        img = Image(0, 0)
        path = args[0]
    elif len(args) == 2:
        img = args[0]
        path = args[1]
    else:
        raise TypeError(
            "load_image(str) or load_image(Image, str)"
        )
    if isinstance(img, Image) and isinstance(path, str):
        lib.load_image(img._id, path.encode("utf-8"))
    return img

def load_font(path: str, scale: float = 1):
    f = Font()
    f._id = lib.load_font(path.encode("utf-8"), 1500 * scale)
    return f



from __future__ import annotations
from ..function import *
import math
from enum import Enum

class Vec2D:
    x: float = 0
    y: float = 0
    def __init__(self, x=0.0, y=0.0):
        self.x = float(x)
        self.y = float(y)

    # Vec2D + Vec2D
    def __add__(self, v: Vec2D):
        return Vec2D(self.x + v.x, self.y + v.y)

    # Vec2D - Vec2D
    def __sub__(self, v: Vec2D):
        return Vec2D(self.x - v.x, self.y - v.y)

    # Vec2D * scalar
    def __mul__(self, s: float):
        if isinstance(s, (int, float)):
            return Vec2D(self.x * s, self.y * s)
        return NotImplemented
    
    # Vec2D / scalar
    def __truediv__(self, s: float):
        if isinstance(s, (int, float)):
            return Vec2D(self.x / s, self.y / s)
        return NotImplemented

    # Vec2D += Vec2D
    def __iadd__(self, v: Vec2D):
        self.x += v.x
        self.y += v.y
        return self
    
    # Vec2D -= Vec2D
    def __isub__(self, v: Vec2D):
        self.x -= v.x
        self.y -= v.y
        return self

    # Vec2D *= scalar
    def __imul__(self, s: float):
        if isinstance(s, (int, float)):
            self.x *= s
            self.y *= s
            return self
        return NotImplemented

    # Vec2D /= scalar
    def __itruediv__(self, s: float):
        if isinstance(s, (int, float)):
            self.x /= s
            self.y /= s
            return self
        return NotImplemented

    # 長さ（距離）
    def length(self):
        return math.sqrt(self.x * self.x + self.y * self.y)

    def __repr__(self):
        return f"Vec2D(x: {self.x}, y: {self.y})"

class Vec3D:
    x: float = 0
    y: float = 0
    z: float = 0
    def __init__(self, x=0.0, y=0.0, z=0.0):
        self.x = float(x)
        self.y = float(y)
        self.z = float(z)

     # Vec3D + Vec3D
    def __add__(self, v: Vec3D):
        return Vec3D(self.x + v.x, self.y + v.y, self.z + v.z)

    # Vec3D - Vec3D
    def __sub__(self, v: Vec3D):
        return Vec3D(self.x - v.x, self.y - v.y, self.z - v.z)
        
    # Vec3D * scalar
    def __mul__(self, s: float):
        if isinstance(s, (int, float)):
            return Vec3D(self.x * s, self.y * s, self.z * s)
        return NotImplemented
    
    # Vec3D / scalar
    def __truediv__(self, s: float):
        if isinstance(s, (int, float)):
            return Vec3D(self.x / s, self.y / s, self.z / s)
        return NotImplemented
        
    # Vec3D += Vec3D
    def __iadd__(self, v: Vec3D):
        self.x += v.x
        self.y += v.y
        self.z += v.z
        return self
    
    # Vec3D -= Vec3D
    def __isub__(self, v: Vec3D):
        self.x -= v.x
        self.y -= v.y
        self.z -= v.z
        return self

    # Vec3D *= scalar
    def __imul__(self, s: float):
        if isinstance(s, (int, float)):
            self.x *= s
            self.y *= s
            self.z *= s
            return self
        return NotImplemented

    # Vec3D /= scalar
    def __itruediv__(self, s: float):
        if isinstance(s, (int, float)):
            self.x /= s
            self.y /= s
            self.z /= s
            return self
        return NotImplemented

    def length(self):
        return math.sqrt(self.x**2 + self.y**2 + self.z**2)

    def normalized(self):
        l = self.length()
        if l == 0:
            return Vec3D(0,0,0)  # ゼロベクトルの場合
        return Vec3D(self.x / l,
                    self.y / l,
                    self.z / l)
    
    def __repr__(self):
        return f"Vec3D(x: {self.x}, y: {self.y}, z: {self.z})"
    
class Rect:
    def __init__(self, x=0, y=0, w=0, h=0):
        self.x = x
        self.y = y
        self.w = w
        self.h = h

class RGBA:
    def __init__(self, r=255, g=255, b=255, a=255):
        self.r = r
        self.g = g
        self.b = b
        self.a = a

class Event:
    def __init__(self):
        self._id = -1

    def pressed(self, key):
        return lib.event_pressed(self._id, key)
    
    def released(self, key):
        return lib.event_released(self._id, key)
    
    def down(self, key):
        return lib.event_down(self._id, key)
        
    def up(self, key):
        return lib.event_up(self._id, key)
    
class SurfaceType(Enum):
    TEXTURE = 1
    WINDOW = 2

class Surface:
    _id: int
    type: SurfaceType
    def __init__(self):
        pass

class Window(Surface):
    # _id: int
    def __init__(self, width, height, title):
        self._id = lib.create_window(int(width), int(height), title.encode("utf-8"), True)
        self.__width = width
        self.__height = height
        self.__title = title
        self.type = SurfaceType.WINDOW

    def closed(self):
        return lib.window_closed(self._id)
    
    def clear(self, color: RGBA):
        lib.clear_win(self._id, color.r / 255.0, color.g / 255.0, color.b / 255.0, color.a / 255.0)

    def swap(self):
        lib.swap_win(self._id)

    def get_event(self):
        event = Event()
        event._id = lib.get_event(self._id)
        return event

class Texture:
    def __init__(self, width: int=-1, height: int=-1):
        self._id = -1
        if width >= 0 and height >= 0:
            self._id = lib.create_texture(width, height)
        self.type = SurfaceType.TEXTURE

    def clear(self, color: RGBA):
        lib.clear_tex(self._id, color.r / 255.0, color.g / 255.0, color.b / 255.0, color.a / 255.0)

    @classmethod
    def from_cropped_image(cls, img: Image, rect: Rect):
        # tex = Texture()
        tex = cls(rect.w, rect.h)
        lib.crop_image_to_texture(img._id, rect.x, rect.y, rect.w, rect.h, tex._id, False)
        return tex

class Image:
    def __init__(self, width: int=-1, height: int=-1):
        self._id = -1
        if width >= 0 and height >= 0:
            self._id = lib.create_image(width, height)

class Font:
    def __init__(self):
        self._id = -1


import sys
import ctypes
from types import SimpleNamespace

from .function import *
from .core import *


def _get_evt_cst(name):
    return ctypes.c_int.in_dll(lib, name)

Key = SimpleNamespace()

initialized = False

def _init():
    global initialized    

    if initialized: return
    lib.initialization()
    for ch in "ABCDEFGHIJKLMNOPQRSTUVWXYZ":
        setattr(Key, ch, _get_evt_cst(f"KEY_{ch}"))
    Key.escape = _get_evt_cst("KEY_ESCAPE")
    Key.space = _get_evt_cst("KEY_SPACE")
    Key.arrow_up = _get_evt_cst("KEY_UP")
    Key.arrow_down = _get_evt_cst("KEY_DOWN")
    Key.arrow_right = _get_evt_cst("KEY_RIGHT")
    Key.arrow_left = _get_evt_cst("KEY_LEFT")

    initialized = True


_init()


import ctypes
import sys
from types import SimpleNamespace

import platform
from pathlib import Path

_here = Path(__file__).resolve().parent

if sys.platform.startswith("win"):
    lib_path = _here / "lib" / "Uniuge.dll"
elif sys.platform == "darwin":
    lib_path = _here / "lib" / "Uniuge.dylib"
else:
    raise RuntimeError("Unsupported platform")

lib = ctypes.CDLL(str(lib_path))

c_int = ctypes.c_int
c_float = ctypes.c_float
c_char = ctypes.c_char_p
c_bool = ctypes.c_bool

# Get time since program start
lib.get_time.restype = c_float
lib.get_time.argtypes = None

# Initialize
lib.initialization.argtypes = None
lib.initialization.restype = None

# Create window
lib.create_window.argtypes = [c_int, c_int, c_char, c_bool]
lib.create_window.restype = c_int

# Call every frame
lib.start_frame.argtypes = None
lib.start_frame.restype = None

# Check if window was closed
lib.window_closed.argtypes = [c_int]
lib.window_closed.restype = c_bool

# Swap window
lib.swap_win.argtypes = [c_int]
lib.swap_win.restype = None

lib.clear_win.argtypes = [c_int, c_float, c_float, c_float, c_float]
lib.clear_win.restype = None

lib.clear_tex.argtypes = [c_int, c_float, c_float, c_float, c_float]
lib.clear_tex.restype = None

lib.draw_rect_win.argtypes = [c_int, c_float, c_float, c_float, c_float, c_float, c_float, c_float, c_float]
lib.draw_rect_win.restype = None

lib.create_texture.argtypes = [c_int, c_int]
lib.create_texture.restype = c_int

lib.load_texture.argtypes = [c_int, c_char, c_bool]
lib.load_texture.restype = None

lib.draw_texture_win.argtypes = [c_int, c_int, c_float, c_float, c_float, c_float, c_float, c_float, c_float, c_float, c_bool, c_bool]
lib.draw_texture_win.restype = None

lib.draw_texture_tex.argtypes = [c_int, c_int, c_float, c_float, c_float, c_float, c_float, c_float, c_float, c_float, c_bool, c_bool]
lib.draw_texture_tex.restype = None

# Event
lib.get_event.argtypes = [c_int]
lib.get_event.restype = c_int

lib.event_pressed.argtypes = [c_int, c_int]
lib.event_pressed.restype = c_bool

lib.event_released.argtypes = [c_int, c_int]
lib.event_released.restype = c_bool

lib.event_down.argtypes = [c_int, c_int]
lib.event_down.restype = c_bool

lib.event_up.argtypes = [c_int, c_int]
lib.event_up.restype = c_bool

lib.get_time.argtypes = None
lib.get_time.restype = ctypes.c_double

lib.get_time.argtypes = None
lib.get_time.restype = ctypes.c_double

lib.load_font.argtypes = [c_char, c_float]
lib.load_font.restype = c_int

lib.draw_text_win.argtypes = [c_int, c_int, c_char, c_float, c_float, c_float, c_float, c_float, c_float, c_float]
lib.draw_text_win.restype = None

lib.draw_text_tex.argtypes = [c_int, c_int, c_char, c_float, c_float, c_float, c_float, c_float, c_float, c_float]
lib.draw_text_tex.restype = None


# -- 3D --
lib.create_space3d.argtypes = None
lib.create_space3d.restype = c_int

lib.draw_camera_3d_win.argtypes = [c_int, c_int, c_float, c_float, c_float, c_float, c_float, c_float, c_int, c_int, c_float, c_float]
lib.draw_camera_3d_win.restype = None

lib.draw_camera_3d_tex.argtypes = [c_int, c_int, c_float, c_float, c_float, c_float, c_float, c_float, c_int, c_int, c_float, c_float]
lib.draw_camera_3d_tex.restype = None

lib.clear_space3d.argtypes = [c_int]
lib.clear_space3d.restype = None

# adding a 3D cube to the 3D space
lib.draw_cube_3d.argtypes = [c_int,
                 c_float,c_float,c_float, # position(x,y,z)
                 c_float,c_float,c_float, # rotation(x,y,z)
                 c_float,c_float,c_float, # size(x,y,z)
                 c_float,c_float,c_float,c_float]
lib.draw_cube_3d.restype = None

lib.draw_ambient_light_3d.argtypes = [c_int, c_float, c_float, c_float, c_float, c_float]
lib.draw_ambient_light_3d.restype = None

lib.draw_point_light_3d.argtypes = [c_int, c_float, c_float, c_float, c_float, c_float, c_float, c_float, c_float]
lib.draw_point_light_3d.restype = None

lib.draw_rect_tex.argtypes = [c_int, c_float, c_float, c_float, c_float, c_float, c_float, c_float, c_float]
lib.draw_rect_tex.restype = None


# -- Image --

# Create image
lib.create_image.argtypes = [c_int, c_int]
lib.create_image.restype = c_int

# Load image
lib.load_image.argtypes = [c_int, c_char]
lib.load_image.restype = None

# 
lib.crop_image_to_texture.argtypes = [c_int, c_int, c_int, c_int, c_int, c_int, c_bool]
lib.crop_image_to_texture.restype = None